@@ -1,23 +1,38 @@
+# This file is part of Buildbot.  Buildbot is free software: you can
+# redistribute it and/or modify it under the terms of the GNU General Public
+# License as published by the Free Software Foundation, version 2.
+#
+# This program is distributed in the hope that it will be useful, but WITHOUT
+# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
+# details.
+#
+# You should have received a copy of the GNU General Public License along with
+# this program; if not, write to the Free Software Foundation, Inc., 51
+# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
+#
+# Copyright Buildbot Team Members
+
 
-import os, signal, types, re, traceback
-from stat import ST_CTIME, ST_MTIME, ST_SIZE
 import os
+from base64 import b64encode
 import sys
 import shutil
 
 from zope.interface import implements
-from twisted.internet import reactor, defer, task
+from twisted.internet import reactor, defer
 from twisted.python import log, failure, runtime
 
 from buildslave.interfaces import ISlaveCommand
 from buildslave import runprocess
 from buildslave.exceptions import AbandonChain
 from buildslave.commands import utils
+from buildslave import util
 
 # this used to be a CVS $-style "Revision" auto-updated keyword, but since I
 # moved to Darcs as the primary repository, this is updated manually each
 # time this file is changed. The last cvs_ver that was here was 1.51 .
-command_version = "2.11"
+command_version = "2.14"
 
 # version history:
 #  >=1.17: commands are interruptable
@@ -45,6 +60,9 @@ command_version = "2.11"
 #  >= 2.9: add depth arg to SVN class
 #  >= 2.10: CVS can handle 'extra_options' and 'export_options'
 #  >= 2.11: Arch, Bazaar, and Monotone removed
+#  >= 2.12: SlaveShellCommand no longer accepts 'keep_stdin_open'
+#  >= 2.13: SlaveFileUploadCommand supports option 'keepstamp'
+#  >= 2.14: RemoveDirectory can delete multiple directories
 
 class Command:
     implements(ISlaveCommand)
@@ -114,6 +132,7 @@ class Command:
         self.builder = builder
         self.stepId = stepId # just for logging
         self.args = args
+        self.startTime = None
         self.setup(args)
 
     def setup(self, args):
@@ -122,8 +141,10 @@ class Command:
 
     def doStart(self):
         self.running = True
+        self.startTime = util.now(self._reactor)
         d = defer.maybeDeferred(self.start)
         def commandComplete(res):
+            self.sendStatus({"elapsed": util.now(self._reactor) - self.startTime})
             self.running = False
             return res
         d.addBoth(commandComplete)
@@ -231,6 +252,7 @@ class SourceBaseCommand(Command):
         self.timeout = args.get('timeout', 120)
         self.maxTime = args.get('maxTime', None)
         self.retry = args.get('retry')
+        self.logEnviron = args.get('logEnviron',True)
         self._commandPaths = {}
         # VC-specific subclasses should override this to extract more args.
         # Make sure to upcall!
@@ -259,12 +281,19 @@ class SourceBaseCommand(Command):
             self.srcdir = self.workdir
 
         self.sourcedatafile = os.path.join(self.builder.basedir,
-                                           ".buildbot-sourcedata")
+            ".buildbot-sourcedata-" + b64encode(self.srcdir))
+
         # upgrade older versions to the new sourcedata location
         old_sd_path = os.path.join(self.builder.basedir, self.srcdir, ".buildbot-sourcedata")
         if os.path.exists(old_sd_path) and not os.path.exists(self.sourcedatafile):
             os.rename(old_sd_path, self.sourcedatafile)
 
+        # also upgrade versions that didn't include the encoded version of the
+        # source directory
+        old_sd_path = os.path.join(self.builder.basedir, ".buildbot-sourcedata")
+        if os.path.exists(old_sd_path) and not os.path.exists(self.sourcedatafile):
+            os.rename(old_sd_path, self.sourcedatafile)
+
         d = defer.succeed(None)
         self.maybeClobber(d)
         if not (self.sourcedirIsUpdateable() and self.sourcedataMatches()):
@@ -343,6 +372,12 @@ class SourceBaseCommand(Command):
         return None
 
     def readSourcedata(self):
+        """
+        Read the sourcedata file and return its contents
+
+        @returns: source data
+        @raises: IOError if the file does not exist
+        """
         return open(self.sourcedatafile, "r").read()
 
     def writeSourcedata(self, res):
@@ -366,6 +401,12 @@ class SourceBaseCommand(Command):
             return rc
         if self.interrupted:
             raise AbandonChain(1)
+
+        # allow AssertionErrors to fall through, for benefit of the tests; for
+        # all other errors, carry on to try the fallback
+        if isinstance(rc, failure.Failure) and rc.check(AssertionError):
+            return rc
+
         # Let VCS subclasses have an opportunity to handle
         # unrecoverable errors without having to clobber the repo
         self.maybeNotDoVCFallback(rc)
@@ -461,7 +502,7 @@ class SourceBaseCommand(Command):
         command = ["rm", "-rf", d]
         c = runprocess.RunProcess(self.builder, command, self.builder.basedir,
                          sendRC=0, timeout=self.timeout, maxTime=self.maxTime,
-                         usePTY=False)
+                         logEnviron=self.logEnviron, usePTY=False)
 
         self.command = c
         # sendRC=0 means the rm command will send stdout/stderr to the
@@ -492,7 +533,7 @@ class SourceBaseCommand(Command):
                                 '-exec', 'chmod', 'u+rwx', '{}', ';' ]
         c = runprocess.RunProcess(self.builder, command, self.builder.basedir,
                          sendRC=0, timeout=self.timeout, maxTime=self.maxTime,
-                         usePTY=False)
+                         logEnviron=self.logEnviron, usePTY=False)
 
         self.command = c
         d = c.start()
@@ -521,7 +562,7 @@ class SourceBaseCommand(Command):
         command = ['cp', '-R', '-P', '-p', fromdir, todir]
         c = runprocess.RunProcess(self.builder, command, self.builder.basedir,
                          sendRC=False, timeout=self.timeout, maxTime=self.maxTime,
-                         usePTY=False)
+                         logEnviron=self.logEnviron, usePTY=False)
         self.command = c
         d = c.start()
         d.addCallback(self._abandonOnFailure)
@@ -559,7 +600,8 @@ class SourceBaseCommand(Command):
         # now apply the patch
         c = runprocess.RunProcess(self.builder, command, dir,
                          sendRC=False, timeout=self.timeout,
-                         maxTime=self.maxTime, usePTY=False)
+                         maxTime=self.maxTime, logEnviron=self.logEnviron,
+                         usePTY=False)
         self.command = c
         d = c.start()
         
@@ -575,3 +617,11 @@ class SourceBaseCommand(Command):
         d.addCallback(self._abandonOnFailure)
         return d
 
+    def setFileContents(self, filename, contents):
+        """Put the given C{contents} in C{filename}; this is a bit more
+        succinct than opening, writing, and closing, and has the advantage of
+        being patchable in tests.  Note that the enclosing directory is
+        not automatically created, nor is this an "atomic" overwrite."""
+        f = open(filename, 'w')
+        f.write(contents)
+        f.close()
